function G = compute_geometry( M, options )

% compute_geometry - compute the geometry for a bandelet transform.
%
% G = compute_geometry( M, options );
%
%   options.T is the threshold for the penalization of the geometry.
%   options.w is the width of the squares.
%   options.q control the overlap. Set q=w for no overlap and q=w/2 for
%       half overlap.
%
%   If M is a 2D array, compute the optimal geometry for each square of
%       width w in M.
%   If M is a cell array, compute the geometry for each entry of the array.
%   If options.callback is a hande of a function of the type
%       [MW, ind] = function(M,options);
%   where MW is a cell array (transformed domain)
%       then the geometry is computed for each MW{i} for i in ind.
%
%   Copyright (c) 2007 Gabriel Peyre


options.null = 0;

if isfield(options, 'T')
    T = options.T;
else
    warning('You should define options.T');
    T = 1;
end
if isfield(options, 'w')
    w = options.w;
else
    w = 8;
end
if isfield(options, 'q')
    q = options.q;
else
    q = w;
end
if isfield(options, 'remove_mean')
    remove_mean = options.remove_mean;
else
    remove_mean = 0;
end

if isfield(options, 'callback')  && not(isempty(options.callback))
    if iscell(M)
        error('Cannot use options.callback for cell arrays.');
    end
    MW = feval(options.callback, M, options);
    options.callback = [];
    if isfield(options, 'Jband')
        Jband = options.Jband;
    else
        Jband = 6;
    end
    G = compute_geometry( {MW{1:Jband}}, options );
    return;
end

if iscell(M)
    G = {};
    for i=1:length(M)
        progressbar(i,length(M));
        G{i} = compute_geometry( M{i}, options );
    end
    return;
end

n = size(M,1);
ww = w^2; % number of pixel in a patch

H = perform_square_extraction(M,w,n,q);
p = size(H,3); % number of patches
if remove_mean
    H = H - repmat( mean(mean(H)), [w w 1] );
end

% compute the LUT for the dictional orderings
L = compute_geometric_lut(w);
m = size(L,2); % number of geometries

% compute the 1D
H = reshape(H, [ww, p]);
L = repmat( reshape(L,[ww 1 m]), [1 p 1]);
L = L + repmat( 0:ww:ww*(p-1), [ww 1 m] ); 

% here are the scrambled signals
HL = H(L);

% perform the 1D haar transform
if isfield(options, 'Jhaar')
    J = options.Jhaar;
else
    J = 3;
end
HW = perform_haar_transform(HL,+1,J, options);
ww0 = size(HW,1);
% add untransformed data
HW = cat( 3, HW, repmat(H, [ww0/ww 1]) );

lagrangian_type = 'tv';
lagrangian_type = 'err';
% lagrangian_type = 'quant';

% price for a geometry coefficient
if isfield(options, 'geometry_weight')
    c = options.geometry_weight;
else
    c = 1;
end


switch lagrangian_type
    case 'err'
        % compute thresholded error
        % compute the number of important coefficients
        if isfield(options, 'ti') && options.ti
            % take into account redundancy of the dictionary
            A = zeros([1 p m+1]);
            for j=1:J+1
                sel = (J+1-j)*ww+1:(J+2-j)*ww;
                mu = 1/2^j;
                if j==J+1
                    mu = 1/2^j;
                end
                A = A + sum( (abs(HW(sel,:,:))>=T), 1  )*mu;
            end
            A(1,:,end) = sum( abs(H)>=T );
            % compute error
            HWT = perform_thresholding(HW, T);
            H1 = perform_haar_transform(HWT,-1,J, options);
            H1(:,:,end) = perform_thresholding(H,T);
            HL(:,:,end+1) = H;
            E = sum( (HL-H1).^2 );
        else
            % error
            E = sum( (HW.^2).*(abs(HW)<T), 1  );
            % traditional orthogonal
            A = sum( (abs(HW)>=T), 1  );
        end
        % add penalty for geometry (1 coef)
        A(:,:,1:end-1) = A(:,:,1:end-1) + c;
        % compute the lagrangian
        L = E+A*T^2;
    case 'tv'
        % use total variation
        L = sum( abs(diff(HW)), 1 );
    case 'quant'
        [HWT, y_quant] = perform_quantization(HW, T, 1);
        E = sum( (HWT-HW).^2, 1 );
        % number of coefficients
        A = sum( (abs(HW)>=T), 1  );
        % add penalty for geometry (1 coef)
        A(:,:,1:end-1) = A(:,:,1:end-1) + c;
        % compute the lagrangian
        L = E + 3/4 * A*T^2;
end

% minimize the lagrandian E+A*T^2
[tmp,G] = min(L,[],3);
if q==w
    G = reshape(G, sqrt([p p]));
end

% special token : no geometry
G(G==m+1) = 0;
