function M = perform_bandlet_transform( M, G, options )

% perform_bandlet_transform - compute the geometry for a bandelet transform.
%
% MB = perform_bandlet_transform( M, G, options );
%
%   options.T is the threshold for the penalization of the geometry.
%   options.w is the width of the squares.
%	options.q controls overlap.
%
%   If M is a 2D array, compute the optimal geometry for each square of
%       width w in M.
%   If M is a cell array, compute the geometry for each entry of the array.
%   If options.callback is a hande of a function of the type
%       [MW, ind] = function(M,options);
%   where MW is a cell array (transformed domain)
%       then the geometry is computed for each MW{i} for i in ind.
%
%   Copyright (c) 2007 Gabriel Peyr?

options.null = 0;
if isfield(options, 'T')
    T = options.T;
else
    warning('You should define options.T');
    T = 1;
end
if isfield(options, 'w')
    w = options.w;
else
    w = 8;
end
if isfield(options, 'q')
    q = options.q;
else
    q = w;
end

if isfield(options, 'callback') && not(isempty(options.callback))
    if iscell(M)
        dir = -1;
    else
        dir = 1;
    end
    cb = options.callback;
    options.callback = [];
    ind = 1:length(G);
    options.dir = dir;
    if dir==1
        MW = feval(cb, M, options);
        M = perform_bandlet_transform( {MW{ind}}, G, options );
        M = {M{:} MW{ind(end)+1:end}};
    else
        MW = perform_bandlet_transform( {M{ind}}, G, options );
        M = { MW{:} M{ind(end)+1:end} };
        M = feval(cb, M, options);
    end
    return;
end


if iscell(M)
    MB = {};
    if not(length(M)==length(G)) || not(iscell(G))
        error('G should be of same size as G.');
    end
    for i=1:length(M)
        MB{i} = perform_bandlet_transform( M{i}, G{i}, options );
    end
    M = MB;
    return;
end

if not(isfield(options, 'dir'))
    error('You should provide options.dir');
end
dir = options.dir;

ww = w^2; % number of pixel in a patch
if dir==1
    n = size(M,1);   
else
    if isfield(options, 'n')
        n = options.n;
    else
        warning('You should provide options.n');
        n = w*size(M,1);
    end
end


% compute the LUT for the dictional orderings
L = compute_geometric_lut(w);
m = size(L,2)+1;        % number of geometries
L(:,end+1) = (1:ww)';   % untransformed
G(G==0) = m;            % untransformed token

% number of scale for the haar transform
if isfield(options, 'Jhaar')
    J = options.Jhaar;
else
    J = 3;
end

if dir==1
    H = perform_square_extraction(M,w,n,q);
    p = size(H,3);
    H = reshape(H, [ww p]);   % set of patches
    for i=1:p
        % reorder coef according to geometry
        H(:,i) = H( L(:,G(i)), i );
    end
    % perform haar transform
    HW = perform_haar_transform(H,+1,J, options);
    ww1 = size(HW,1);
    % re-assign untransformed data
    HW(:,G==m) = repmat( H(:,G==m), [ww1/ww 1]);
    M = HW;
else
    HW = M;
    p = size(HW,2); % number of patches
    ww1 = size(HW,1); % number of coeff
    % inverse transform
    H = perform_haar_transform(HW,-1,J, options);
    H(:,G==m) = HW(1:ww,G==m);
    % reassign coefs
    for i=1:p
        % reorder coef according to geometry
        H( L(:,G(i)), i ) = H(:,i);
    end
    % reconstruct from squares
    H = reshape(H, [w w p]); 
    M = perform_square_extraction(H,w,n,q);
end