function H = perform_square_extraction(M,w,n,q)

% perform_square_extraction - extract squares from an image
%
% Extraction:
%   H = perform_square_extraction(M,w,n,q);
% Reconstruction:
%   M = perform_square_extraction(H,w,n,q);
%
%   H(:,:,i) is the ith patch of size (w,w) extracted from M.
%
%   q controls overlap:
%       q=w : no overlap
%       q=w/2 : 50% overlap (x4 redundancy)
%
%   Copyright (c) 2007 Gabriel Peyre

if nargin<4
    q = w;
end
if size(M,3)>1
    dir = -1;
else
    dir = 1;
end
if dir==1 && not(n==size(M,1))
    warning('n should be the size of the image');
    n = size(M,1);
end
if dir==-1 && not(w==size(M,1))
    warning('w should be the size of the coefficients');
    w = size(M,1);
end

% cut the image into square of size w
x = 1:q:n-w+1;
[Y,X] = meshgrid(x,x); % corner of the squares
p = size(X(:),1); % number of square
[dY,dX] = meshgrid(0:w-1,0:w-1);
X = repmat( reshape(X,[1,1,p]) ,[w w 1]) + repmat(dX,[1 1 p]);
Y = repmat( reshape(Y,[1,1,p]) ,[w w 1]) + repmat(dY,[1 1 p]);

if max(X(:))<n
    % Here comes the tricky part ...
    % ... handle special case where w does not divide n ...
    % squares are missing on the boundary, add them
    y = 1:w:n-w+1;
    x = (n-w+1)*(y*0+1);
    p = length(x);
    X1 = repmat( reshape(x,[1,1,p]) ,[w w 1]) + repmat(dX,[1 1 p]);
    Y1 = repmat( reshape(y,[1,1,p]) ,[w w 1]) + repmat(dY,[1 1 p]);
    X = cat(3,X,X1);
    Y = cat(3,Y,Y1);
    % squares are missing on the boundary, add them
    x = 1:w:n-w+1;
    y = (n-w+1)*(x*0+1);
    p = length(x);
    X1 = repmat( reshape(x,[1,1,p]) ,[w w 1]) + repmat(dX,[1 1 p]);
    Y1 = repmat( reshape(y,[1,1,p]) ,[w w 1]) + repmat(dY,[1 1 p]);
    X = cat(3,X,X1);
    Y = cat(3,Y,Y1);
    % squares are missing on the boundary, add them
    x = n-w+1;
    y = n-w+1;
    p = length(x);
    X1 = repmat( reshape(x,[1,1,p]) ,[w w 1]) + repmat(dX,[1 1 p]);
    Y1 = repmat( reshape(y,[1,1,p]) ,[w w 1]) + repmat(dY,[1 1 p]);
    X = cat(3,X,X1);
    Y = cat(3,Y,Y1);
end

% compute indexes
I = sub2ind([n n], X,Y);
if dir==1
    % extraction
    H = M(I);
else
    % reconstruction
    Weight = zeros(n);
    H = zeros(n);
    for i=1:size(I,3)
        Weight(I(:,:,i)) =  Weight(I(:,:,i)) + 1;
        H(I(:,:,i)) = H(I(:,:,i)) + M(:,:,i);
    end
    if sum(Weight(:)==0)>1
        warning('Problem with reconstruction');
    end
    Weight(Weight==0)=1;
    H = H ./ Weight;
end